/* Thread termination.
   Copyright (C) 2000 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the GNU C Library; see the file COPYING.LIB.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>

#include <pt-internal.h>

#include <bits/atomic.h>
#if 0
#include <hurd/ihash.h>

#define __ihash_find	ihash_find
#define __ihash_free	ihash_free
#define __ihash_iterate	ihash_iterate
#define __ihash_remove  ihash_remove

/* Call the cancellation cleanup handlers on the list CLEANUP.  */
static void
do_cleanup (struct __pthread *thread)
{
  int oldstate;

  /* According to POSIX, the cancellation cleanup handlers
     should be called with cancellation disabled.  */
  pthread_setcancelstate (PTHREAD_CANCEL_DISABLE, &oldstate);

  do
    {
      thread->cleanup->__routine (thread->cleanup->__arg);
      thread->cleanup = thread->cleanup->__next;
    }
  while (thread->cleanup);

  pthread_setcancelstate (oldstate, NULL);
}


/* Table of key destructor functions.  */
pthread_key_destructor_t *__pthread_key_destructors = NULL;
int __pthread_num_keys = 0;

/* XXX */
#define __PTHREAD_RECURSIVE_MUTEX_INITIALIZER PTHREAD_MUTEX_INITIALIZER

/* Lock for `__pthread_key_destructors'.  Note that the destructors are
   allowed to call pthread_key_delete!.  That's why this is a
   recursive mutex.  */
pthread_mutex_t __pthread_key_lock = __PTHREAD_RECURSIVE_MUTEX_INITIALIZER;

/* List of thread-specific data hashes that are in use.  */
struct pthread_specific *__pthread_specific = NULL;
__spin_lock_t __pthread_specific_lock = __SPIN_LOCK_INITIALIZER; 

/* Call the thread-specific data destructor functions for SPECIFIC.  */
static void internal_function
destroy_specific (struct pthread_specific *specific)
{
  void *value;
  int i;

  error_t check (void *value)
    {
      /* Just a random error value, that feels appropriate.  */
      return EAGAIN;
    }
  
  /* Remove hash from the list of thread-specific data hashes that are
     in use.  */
  __spin_lock (&__pthread_specific_lock);
  if (specific->next)
    specific->next->prevp = specific->prevp;
  *specific->prevp = specific->next;
  __spin_unlock (&__pthread_specific_lock);

  /* Iterate over all thread-sepcific data keys, calling the
     destructor functions for all non-NULL values with an associated
     destructor.  Before calling the destructor, the value is reset to
     NULL (by removing the value from the hash table).  This is not
     mentioned in the official standard, but according to a defect
     report this was the intended behaviour.  */

  /* We lock the table of destructor functions to prevent
     `pthread_key_delete' from removing a key while we're executing
     the associated destructor.  */
  __pthread_mutex_lock (&__pthread_key_lock);
  while (__ihash_iterate (specific->ht, check) == EAGAIN)
    {
      for (i = 0; i < __pthread_num_keys; i++)
	{
	  if (__pthread_key_destructors[i] != PTHREAD_KEY_AVAILABLE &&
	      __pthread_key_destructors[i] != NULL &&
	      (value = __ihash_find (specific->ht, i)))
	    {
	      __ihash_remove (specific->ht, i);
	      __pthread_key_destructors[i] (value);
	    }
	}
    }
  __pthread_mutex_unlock (&__pthread_key_lock);

  __ihash_free (specific->ht);
  free (specific);
}
#endif


/* Terminate the current thread and make STATUS available to any
   thread that might join us.  */
void
pthread_exit (void *status)
{
  struct __pthread *self = _pthread_self ();
  struct __pthread_cancelation_handler **handlers;

  /* Run any cancelation handlers.  */
  for (handlers = __pthread_get_cleanup_stack ();
       *handlers;
       *handlers = (*handlers)->next)
    (*handlers)->handler ((*handlers)->arg);

#if 0
  /* Call cleanup handlers.  */
  if (self->cleanup)
    do_cleanup (self);
  
  if (self->specific)
    destroy_specific (self->specific);
#endif

  /* Decrease the number of threads.  We use an atomic operation to
     make sure that only the last thread calls `exit'.  */
  if (__atomic_dec_and_test (&__pthread_total))
    /* We are the last thread.  */
    exit (0);

  /* Note that after this point the process can be terminated at any
     point if another thread calls `pthread_exit' and happens to be
     the last thread.  */

  __pthread_mutex_lock (&self->state_lock);

  if (self->cancel_state == PTHREAD_CANCEL_ENABLE && self->cancel_pending)
    status = PTHREAD_CANCELED;

  switch (self->state)
    {
    default:
      assert (! "This cannot happen!");
      break;

    case PTHREAD_DETACHED:
      /* Make sure that nobody can reference this thread anymore, and
         mark it as terminated.  Our thread ID will immediately become
         available for re-use.  For obvious reasons, we cannot
         deallocate our own stack.  However, it will eventually be
         reused when this thread structure is recycled.  */
      __pthread_mutex_unlock (&self->state_lock);
      __pthread_dealloc (self);

      break;

    case PTHREAD_JOINABLE:
      /* We need to stay around for a while since another thread
         might want to join us.  */
      self->state = PTHREAD_EXITED;

      /* We need to remember the exit status.  A thread joining us
         might ask for it.  */
      self->status = status;

      /* Broadcast the condition.  This will wake up threads that are
         waiting to join us.  */
      pthread_cond_broadcast (&self->state_cond);
      __pthread_mutex_unlock (&self->state_lock);

      break;
    }

  /* Note that after this point the resources used by this thread can
     be freed at any moment if another thread joins or detaches us.
     This means that before freeing any resources, such a thread
     should make sure that this thread is really halted.  */
  
  __pthread_halt (self);

  /* NOTREACHED */
  abort ();
}
